/*
 * cc2511.c
 */

#include "cc2511.h"

/**************************************************************************//**
* @brief    Resets the DUP into debug mode. Function assumes that
*           the programmer I/O has already been configured using e.g.
*           programmer_init().
*
* @return   None.
******************************************************************************/
void cc2511_debug_init(void)
{
    /* Toggle the pins */
	RST_HIGH;
	DC_LOW;
	DD_LOW;
	/* Give cycles to let everything settle */
	DELAY(5);

	/****************************************************************
	 * Put DUP into debug mode
	 * 	1. Pull RST down
	 * 	2. Two clock pulses
	 * 	3. Pull RST up
	 */

	/* 1. Send two flanks on DC while keeping RESET_N low */
	RST_LOW;
	DELAY(4);

	/* 2. Two clock cycles */

	/* First clock cycle */
	DC_LOW;
	DELAY(1);
	DC_HIGH;
	DELAY(1);

	/* Second clock cycle */
	DC_LOW;
	DELAY(1);
	DC_HIGH;
	DELAY(1);

	/* Keep the clock low */
	DC_LOW;

	/* 	3. Pull RST up */
	DELAY(5);
	RST_HIGH;
}

/**************************************************************************//**
* @brief    Start the 32Mhz clock on the DUP. Function assumes that
*           the DUP has been halted.
*
* @return   None.
******************************************************************************/
void cc2511_clock_init(void)
{
	uint8_t status = 0x00;
    uint8_t clock_instr[] = {0x75, 0xC6, 0x00};
    uint8_t status_instr[] = {0xE5, 0xBE};

    // MOV DPTR, address
    cc2511_debug_command(CMD_DEBUG_INSTR_3B, clock_instr, 3);

    /* Give a clock cycle for the next command */
    DELAY(1);

    cc2511_debug_command(CMD_DEBUG_INSTR_2B, status_instr, 2);

    /* Give a clock cycle for the next command */
    DELAY(1);

    /*
     * Wait for status bit 1 to go high
     * DS page 72, table 47:
     * Bit	Field Name			Description
     * -------------------------------------------------
     * 1 	OSCILLATOR_STABLE 	Oscillators stable.
     * 							This bit represents the status of the SLEEP.XSOC_STB
     * 							and SLEEP.HFRC_STB register bits.
     * 							0 Oscillators not stable
     * 							1 Oscillators stable
     * -------------------------------------------------
     */
    do{
    	status = cc2511_read_status();
    } while ((status & STATUS_OSC_STABLE_BM) != 0x02);

    return;
}

/**************************************************************************//**
* @brief    Read the Status over the debug interface using the
*           Status command.
*
* @return   None
******************************************************************************/
uint8_t cc2511_read_status(void)
{
	unsigned char output = 0x00;

	// Make sure DD is output
    SET_DD_OUTPUT;
    /* Send Status Command */
    cc2511_write_debug_byte(CMD_READ_STATUS);

    // Make sure DD is output
    SET_DD_INPUT;
    // Wait for data to be ready
    cc2511_wait_dup_ready();

    output = cc2511_read_debug_byte();

    return output;
}

/**************************************************************************//**
* @brief    Issues a CHIP_ERASE command on the debug interface and waits for it
*           to complete.
*
* @return   None.
******************************************************************************/
void cc2511_chip_erase(void)
{
	volatile unsigned char status;
    // Send command
    cc2511_debug_command(CMD_CHIP_ERASE, 0, 0);

    /*
     * Wait for status bit 7 to go high
     * DS page 72, table 47:
     * Bit	Field Name			Description
     * -------------------------------------------------
     * 7 	CHIP_ERASE_DONE 	Flash chip erase done
     * 							0 Chip erase in progress
     * 							1 Chip erase done
     * -------------------------------------------------
     */
    do {
        status = cc2511_debug_command(CMD_READ_STATUS, 0, 0);
    } while((status & STATUS_CHIP_ERASE_BUSY_BM));
}

/**************************************************************************//**
* @brief    Halts the CPU over the debug interface using the
*           HALT command.
*
* @return   None
******************************************************************************/
void cc2511_halt_cpu(void)
{
    SET_DD_OUTPUT;

    /* Send HALT Command */
    cc2511_write_debug_byte(0x44);

    //Read the dummy byte that Halt sends
    cc2511_read_debug_byte();

    return;
}

/**************************************************************************//**
* @brief    Resumes the CPU over the debug interface using the
*           RESUME command.
*
* @return   None
******************************************************************************/
void cc2511_resume_cpu(void)
{
    SET_DD_OUTPUT;

    /* Send HALT Command */
    cc2511_write_debug_byte(0x4C);

    //Read the dummy byte that Resume sends
    cc2511_read_debug_byte();

    return;
}

/**************************************************************************//**
* @brief    Reads the chip ID over the debug interface using the
*           GET_CHIP_ID command.
*
* @return   Returns the chip id returned by the DUP
******************************************************************************/
uint16_t cc2511_read_chip_id(void)
{
    unsigned char id, revision = 0;

    // Make sure DD is output
    SET_DD_OUTPUT;

    // Send command
    cc2511_write_debug_byte(CMD_GET_CHIP_ID);

    // Set DD as input
    SET_DD_INPUT;

    // Wait for data to be ready
    cc2511_wait_dup_ready();

    // Read ID and revision
    id = cc2511_read_debug_byte(); 		// ID
    revision = cc2511_read_debug_byte();   // Revision (discard)

    /*
     * Explanation:
     * 1. Shift id << 8 bits, moving it to the MSB, resulting in 0x9100
     * 2. OR the revision to make it the LSB, resulting in 0x9104
     */
    return ((unsigned short)id << 8) | revision;
}

/**************************************************************************//**
* @brief    Sends a block of data over the debug interface using the
*           BURST_WRITE command.
*
* @param    src         Pointer to the array of input bytes
* @param    num_bytes   The number of input bytes
*
* @return   None.
******************************************************************************/
void cc2511_burst_write_block(uint8_t *src, uint8_t num_bytes)
{
    int i;

    // Make sure DD is output
    SET_DD_OUTPUT;

    cc2511_write_debug_byte(CMD_BURST_WRITE | HIBYTE(num_bytes));
    cc2511_write_debug_byte(LOBYTE(num_bytes));
    for (i = 0; i < num_bytes; i++)
    {
        cc2511_write_debug_byte(src[i]);
    }

    // Set DD as input
    SET_DD_INPUT;

    // Wait for DUP to be ready
    //cc2511_wait_dup_ready();

    cc2511_read_debug_byte(); // ignore output

    // Set DD as output
    SET_DD_OUTPUT;
}

/**************************************************************************//**
* @brief    Writes a byte on the debug interface. Requires DD to be
*           output when function is called.
*
* @param    data    Byte to write
*
* @return   None.
******************************************************************************/
void cc2511_write_debug_byte(uint8_t data)
{
	/* Send command */
	SET_DD_OUTPUT;

	for (int i = 7; i >= 0; i--) // Iterate from MSB to LSB
	{
		/* Toggle the clock line high to signal that the bit is ready to be read */
		DC_HIGH;
		/* Delay to move the rising edge toward the middle of the pulse */
		DELAY(0.5);

		/* Send the data bit on the DD */
		if (data & (1 << i)) {
			/* Set the data high (1) */
			DD_HIGH;
		} else {
			/* Set the data low (0) */
			DD_LOW;
		}

		/* Delay to move the falling edge toward the middle of the pulse */
		DELAY(0.5);
		/* Toggle the clock line low to finish the pulse */
		DC_LOW;
		/* Finish the clock pulse to signal the next bit */
		DELAY(1);
	}

	/* Delay one clock cycle to make analysis easier */
	DELAY(1);
}

/**************************************************************************//**
* @brief    Reads a byte from the debug interface. Requires DD to be
*           input when function is called.
*
* @return   Returns the byte read.
******************************************************************************/
uint8_t cc2511_read_debug_byte(void)
{
	unsigned char data;

	//SET_DD_INPUT

	for (int i = 7; i >= 0; i--) // Read data MSB first
	{
		/* Set clock high to start the cycle */
		DC_HIGH;
		/* Delay to move the rising edge toward the middle of the pulse */
		DELAY(0.5);

		// Read the data bit and shift it into position
		//if (HAL_GPIO_ReadPin(CC2511_DD_GPIO_Port, CC2511_DD_Pin) == GPIO_PIN_SET) {
		if (IS_DD_HIGH) {
		  data |= (1 << i); // Set the bit if the data line is high
		}

		/* Delay to move the falling edge toward the middle of the pulse */
		DELAY(0.5);

		/* Generate a low edge */
		DC_LOW;
		/* Delay to allow data to stabilize on the falling edge */
		DELAY(1);
	}

	/* Delay one clock cycle to make analysis easier */
	DELAY(1);

    return data;
}

/**************************************************************************//**
* @brief    Function waits for DUP to indicate that it is ready. The DUP will
*           pull the DD line low when it is ready. Requires DD to be input when
*           function is called.
*
* @return   Returns 0 if function timed out waiting for DD line to go low
* @return   Returns 1 when DUP has indicated it is ready.
******************************************************************************/
uint8_t cc2511_wait_dup_ready(void)
{
	SET_DD_INPUT;
    // DUP pulls DD low when ready
    unsigned char count = 0;
    while (IS_DD_HIGH && count < 16)
    {
        // Clock out 8 bits before checking if DD is low again
        cc2511_read_debug_byte();
        count++;
    }

    return (count == 16) ? 0 : 1;
}

/**************************************************************************//**
* @brief    Issues a command on the debug interface. Only commands that return
*           one output byte are supported.
*
* @param    cmd             Command byte
* @param    cmd_bytes       Pointer to the array of data bytes following the
*                           command byte [0-3]
* @param    num_cmd_bytes   The number of data bytes (input to DUP) [0-3]
*
* @return   Data returned by command
******************************************************************************/
uint8_t cc2511_debug_command(uint8_t cmd, uint8_t *cmd_bytes, uint8_t num_cmd_bytes)
{
	unsigned short i;
    unsigned char output = 0;

    // Make sure DD is output
    SET_DD_OUTPUT;

    // Send command
    cc2511_write_debug_byte(cmd);

    // Send bytes
    for (i = 0; i < num_cmd_bytes; i++)
    {
        cc2511_write_debug_byte(cmd_bytes[i]);
    }

    // Wait for data to be ready
    cc2511_wait_dup_ready();

    // Read returned byte
    output = cc2511_read_debug_byte();

    // Set DD as output
    SET_DD_OUTPUT;

    return output;
}



/**************************************************************************//**
* @brief    Writes a byte to a specific address in the DUP's XDATA space.
*
* @param    address     XDATA address
* @param    value       Value to write
*
* @return   None.
******************************************************************************/
void cc2511_write_xdata_memory(unsigned short address, unsigned char value)
{
	unsigned char instr[3];

    // MOV DPTR, address
    instr[0] = 0x90;
    instr[1] = address >> 8;
    instr[2] = address;
    cc2511_debug_command(CMD_DEBUG_INSTR_3B, instr, 3);

    instr[0] = 0x74;
    instr[1] = value;
    cc2511_debug_command(CMD_DEBUG_INSTR_2B, instr, 2);

    instr[0] = 0xF0;
    cc2511_debug_command(CMD_DEBUG_INSTR_1B, instr, 1);

    instr[0] = 0xA3;
    cc2511_debug_command(CMD_DEBUG_INSTR_1B, instr, 1);
}

/**************************************************************************//**
* @brief    Writes a block of data to the DUP's XDATA space.
*
* @param    address     XDATA start address
* @param    values      Pointer to the array of bytes to write
* @param    num_bytes   Number of bytes to write
*
* @return   None.
******************************************************************************/
void cc2511_write_xdata_memory_block(int address, uint8_t *values, uint8_t num_bytes)
{
    unsigned char instr[3];
    unsigned short i;

    // MOV DPTR, address
    instr[0] = 0x90;
    instr[1] = HIBYTE(address);
    instr[2] = LOBYTE(address);
    cc2511_debug_command(CMD_DEBUG_INSTR_3B, instr, 3);

    for (i = 0; i < num_bytes; i++)
    {
        // MOV A, values[i]
        instr[0] = 0x74;
        instr[1] = values[i];
        cc2511_debug_command(CMD_DEBUG_INSTR_2B, instr, 2);

        // MOV @DPTR, A
        instr[0] = 0xF0;
        cc2511_debug_command(CMD_DEBUG_INSTR_1B, instr, 1);

        // INC DPTR
        instr[0] = 0xA3;
        cc2511_debug_command(CMD_DEBUG_INSTR_1B, instr, 1);
    }
}

/**************************************************************************//**
* @brief    Read a byte from a specific address in the DUP's XDATA space.
*
* @param    address     XDATA address
*
* @return   Value read from XDATA
******************************************************************************/
uint8_t cc2511_read_xdata_memory(unsigned short address)
{
    unsigned char instr[3];

    // MOV DPTR, address
    instr[0] = 0x90;
    instr[1] = HIBYTE(address);
    instr[2] = LOBYTE(address);
    cc2511_debug_command(CMD_DEBUG_INSTR_3B, instr, 3);

    // MOVX A, @DPTR
    instr[0] = 0xE0;
    return cc2511_debug_command(CMD_DEBUG_INSTR_1B, instr, 1);
}

/**************************************************************************//**
* @brief    Writes 4-2048 bytes to DUP's flash memory. Parameter \c num_bytes
*           must be a multiple of 4.
*
* @param    src         Pointer to programmer's source buffer (in XDATA space)
* @param    start_addr  FLASH memory start address [0x0000 - 0x7FFF]
* @param    num_bytes   Number of bytes to transfer [4-1024]
*
* @return   None.
******************************************************************************/
void cc2511_write_flash_memory_block(uint8_t *src, uint32_t start_addr, uint8_t num_bytes)
{
	//! DUP DMA descriptor
	uint8_t dma_desc_0[8] =
	{
		// Debug Interface -> Buffer
		HIBYTE(DUP_DBGDATA),            // src[15:8]
		LOBYTE(DUP_DBGDATA),            // src[7:0]
		HIBYTE(ADDR_BUF0),              // dest[15:8]
		LOBYTE(ADDR_BUF0),              // dest[7:0]
		0,                              // len[12:8] - filled in later
		0,                              // len[7:0]
		31,                             // trigger: DBG_BW
		0x11                            // increment destination
	};

	//! DUP DMA descriptor
	uint8_t dma_desc_1[8] =
	{
		// Buffer -> Flash controller
		HIBYTE(ADDR_BUF0),              // src[15:8]
		LOBYTE(ADDR_BUF0),              // src[7:0]
		HIBYTE(DUP_FWDATA),             // dest[15:8]
		LOBYTE(DUP_FWDATA),             // dest[7:0]
		0,                              // len[12:8] - filled in later
		0,                              // len[7:0]
		18,                             // trigger: FLASH
		0x42,                           // increment source
	};

    // 1. Write the 2 DMA descriptors to RAM
    cc2511_write_xdata_memory_block(ADDR_DMA_DESC_0, dma_desc_0, 8);
    cc2511_write_xdata_memory_block(ADDR_DMA_DESC_1, dma_desc_1, 8);

    // 2. Update LEN value in DUP's DMA descriptors
    uint8_t len[2] = {HIBYTE(num_bytes), LOBYTE(num_bytes)};
    cc2511_write_xdata_memory_block((ADDR_DMA_DESC_0+4), len, 2);  // LEN, DBG => ram
    cc2511_write_xdata_memory_block((ADDR_DMA_DESC_1+4), len, 2);  // LEN, ram => flash

    // 3. Set DMA controller pointer to the DMA descriptors
    cc2511_write_xdata_memory(DUP_DMA0CFGH, HIBYTE(ADDR_DMA_DESC_0));
    cc2511_write_xdata_memory(DUP_DMA0CFGL, LOBYTE(ADDR_DMA_DESC_0));
    cc2511_write_xdata_memory(DUP_DMA1CFGH, HIBYTE(ADDR_DMA_DESC_1));
    cc2511_write_xdata_memory(DUP_DMA1CFGL, LOBYTE(ADDR_DMA_DESC_1));

    // 4. Set Flash controller start address (wants 16MSb of 18 bit address)
    cc2511_write_xdata_memory(DUP_FADDRH, HIBYTE( (start_addr>>2) ));
    cc2511_write_xdata_memory(DUP_FADDRL, LOBYTE( (start_addr>>2) ));

    // 5. Arm DBG=>buffer DMA channel and start burst write
    cc2511_write_xdata_memory(DUP_DMAARM, CH_DBG_TO_BUF0);
    cc2511_burst_write_block(src, num_bytes);

    // 6. Start programming: buffer to flash
    cc2511_write_xdata_memory(DUP_DMAARM, CH_BUF0_TO_FLASH);
    cc2511_write_xdata_memory(DUP_FCTL, 0x06);

    // 7. Wait until flash controller is done

    uint8_t status = 0x00;
    do {
		cc2511_write_debug_byte(0x34);
		/* Give a clock cycle for the next command */
		DELAY(1);
		status = cc2511_read_debug_byte();
	} while(!((status) & 0x80));

    //while (cc2511_read_xdata_memory(XREG_TO_INT(DUP_FCTL)) & 0x80);
    HAL_Delay(1);
}

/**************************************************************************//**
* @brief    Reads 1-32767 bytes from DUP's flash to a given buffer on the
*           programmer.
*
* @param    bank        Flash bank to read from [0-7]
* @param    address     Flash memory start address [0x0000 - 0x7FFF]
* @param    values      Pointer to destination buffer.
*
* @return   None.
******************************************************************************/
void cc2511_read_flash_memory_block(unsigned char bank,unsigned short flash_addr, unsigned short num_bytes, unsigned char *values)
{
    unsigned char instr[3];
    unsigned short i;
    unsigned short xdata_addr = (0x8000 + flash_addr);

    // 1. Map flash memory bank to XDATA address 0x8000-0xFFFF
    cc2511_write_xdata_memory(DUP_MEMCTR, bank);

    // 2. Move data pointer to XDATA address (MOV DPTR, xdata_addr)
    instr[0] = 0x90;
    instr[1] = HIBYTE(xdata_addr);
    instr[2] = LOBYTE(xdata_addr);
    cc2511_debug_command(CMD_DEBUG_INSTR_3B, instr, 3);

    for (i = 0; i < num_bytes; i++)
    {
        // 3. Move value pointed to by DPTR to accumulator (MOVX A, @DPTR)
        instr[0] = 0xE0;
        values[i] = cc2511_debug_command(CMD_DEBUG_INSTR_1B, instr, 1);

        // 4. Increment data pointer (INC DPTR)
        instr[0] = 0xA3;
        cc2511_debug_command(CMD_DEBUG_INSTR_1B, instr, 1);

        DELAY(1);
    }
}
